#if T_TUNE
#include	<windows.h>
#include	<stdio.h>
#include	<string.h>
#include	"x1.h"
#include	"x1_ctc.h"

#include	"x1_irq.h"

#include	"dosio.h"
#include	"trace.h"


/***********************************************************************
	CTC
***********************************************************************/


typedef struct {
	BYTE	device;		/* device number of daisy chain   */
	BYTE	value[4];	/* command value          */
	WORD	count[4];	/* time counstant (1-256) */
	EV_TIME basetime[4];/* interval of count down */
	EVENT	event[4];	/* event timer     */

} CTC_TABLE;

		CTC_TABLE	CTC[3];
extern	Z80_Regs	R;

static const char *ctc_event_name[3*4] =
{
	"CTC0.0","CTC0.1","CTC0.2","CTC0.3",
	"CTC1.0","CTC1.1","CTC1.2","CTC1.3",
	"CTC2.0","CTC2.1","CTC2.2","CTC2.3"
};

/* event callback: timer overflow */
static void ctc_int_callback(int device)
{
	X1_set_irq_line((BYTE)device);
}

void init_ctc(void)
{
	int chip,ch;
	ZeroMemory(CTC, sizeof(CTC));

	CTC[0].device = DEVICE_CTC0;
	CTC[1].device = DEVICE_CTC1;
	CTC[2].device = DEVICE_CTC2;
	for(chip=0; chip<=2; chip++)
	{
		for(ch=0;ch<4;ch++)
		{
			CTC[chip].value[ch] = 0x03;
			/* callback parmeter is IRQ device number */
			CTC[chip].event[ch].name = ctc_event_name[chip*4+ch];
			CTC[chip].event[ch].callback_param = CTC[chip].device+3-ch;
		}
	}
}

/* update interval time and event timer */
static void ctc_update(CTC_TABLE *ctc,BYTE ch)
{
	EV_TIME interval,clkin,basetime =0;

	BYTE value = ctc->value[ch];

	/* calcrate input source clock */
	if( !(value & 0x02) )
	{
		switch(ch)
		{
		case 1:
		case 2:
			clkin = TIME_IN_HZ(2000000);// CH1,2 : 2MHz
			break;
		case 3:
			clkin = ctc->basetime[0] * ctc->count[0];	// CH3 <- ZC/TO0
			break;
		default:
			//CH.0͂ubbڑȂ̂œ삵Ȃ
			clkin = 0;
		}
		if(value & 0x40)
		{	/* counter mode */
			basetime = clkin;
		}
		else
		{	/* timer mode */
			if( !(value&0x08) || clkin)
			{
				//^C}[\[X͂botxɂ͈ˑASlgŒI
				basetime = TIME_IN_HZ(4000000) * (value&0x20 ? 256:16);
			}
		}
	}

	ctc->basetime[ch] = basetime;

	/* output interval time */
	interval = basetime * ctc->count[ch];

	if(interval == 0)
	{	/* stop */
		event_remove(&ctc->event[ch]);
	}
	else
	{
		if( value & 0x80)
		{	/* with interrupt */
			ctc->event[ch].callback = ctc_int_callback;
		}
		else
		{	/* without interrupt */
			ctc->event[ch].callback = NULL;

			/* clear pending irq */
			X1_clear_irq_line(ctc->device+3-ch);
			/* I[o[wbh炷ׁAPbt߂֔{  */
			//interval *= (TIME_IN_SEC(1) / interval);
		}
		/* Cxgd| */
		ctc->event[ch].interval = interval;
		event_set(&ctc->event[ch],interval);
	}
	/* if CH0 then recalc CH3 because cascade connection */
	if(ch==0)
		ctc_update(ctc,3);
}

//----------------------------------------------------------------------

CTC_TABLE *ctctablep(WORD port) {

	port &= 0xfffc;
	if (port == 0x1fa0) {
		return(CTC);
	}
	if (port == 0x1fa8) {
		return(&CTC[1]);
	}
	if ((x1flg.SOUND_SW) && (port == 0x0704)) {
		return(&CTC[2]);
	}
	return(NULL);
}

X1_IOW x1_ctc_w(WORD port, BYTE value) {

	CTC_TABLE	*ctc;
	WORD		p = port;

	if ((ctc = ctctablep(port)) == NULL) {
		return;
	}
	port &= 0x0003;

	if (ctc->value[port] & 0x04)
	{	/* time constant */
		ctc->count[port] = value ? value : 0x100;
		ctc->value[port] &= (BYTE)~0x06;
		ctc_update(ctc,(BYTE)port);
	}
	else if (value & 1)
	{	/* mode */
		if( ctc->value[port] != value)
		{
			ctc->value[port] = value;
			ctc_update(ctc,(BYTE)port);
		}
	}
	else
	{	/* vector */
		value &= 0xf8;
		X1_set_irq_vector(ctc->device+3,value  );
		X1_set_irq_vector(ctc->device+2,value+2);
		X1_set_irq_vector(ctc->device+1,value+4);
		X1_set_irq_vector(ctc->device+0,value+6);
	}
}

X1_IOR x1_ctc_r(WORD port) {

	CTC_TABLE	*ctc;
	EV_TIME timeleft;

	if ((ctc = ctctablep(port)) == NULL) {
		return 0xff;
	}

	port &= 3;
	if( !event_enabled(&ctc->event[port]) )
		return (BYTE)ctc->count[port];

	/* return 'now' counter by event timer */
	timeleft = event_timeleft(&ctc->event[port]);
	return (BYTE)((timeleft/ctc->basetime[port]) % ctc->count[port] +1);
}

#else //T_TUNE

#include	<windows.h>
#include	<stdio.h>
#include	<string.h>
#include	"x1.h"
#include	"x1_ctc.h"

#include	"dosio.h"
#include	"trace.h"

/***********************************************************************
	CTC
***********************************************************************/

		CTC_TABLE	CTC[3];
extern	Z80_Regs	R;


void init_ctc(void) {

	int		i, j;

	ZeroMemory(CTC, sizeof(CTC));
	for (j=0; j<3; j++) {
		for (i=0; i<4; i++) {
			CTC[j].value[i] = 0x83;
		}
	}
}


void x1_ctc_int(void) {

	CTC_TABLE	*ctc;
	WORD		clk0, clk4, clk2;
	WORD		subcnt;
	int			i, j;
	BYTE		bit;
	BYTE		ctcint_flg;

	ctcint_flg = Z80_Able_interrupt();

	clk4 = x1flg.HSYNC_CLK;
	clk2 = clk4 >> 1;

	if (x1flg.ROM_TYPE >= 2) {
		ctc = CTC;
		j = 0;
	}
	else {
		ctc = &CTC[2];
		j = 2;
	}

	for (; j<3; j++, ctc++) {
		clk0 = 0;
		bit = 1;
		for (i=0; i<4; i++, bit<<=1) {
			if (!(ctc->value[i] & 0x02)) {
				if (ctc->count[i] <= 0L) {
					ctc->count[i] += ctc->countmax[i];
				}
				if (!(ctc->value[i] & 0x40)) {
					subcnt = clk4;
				}
				else if (i == 3) {
					subcnt = clk0;
				}
				else {
					subcnt = clk2;
				}
				ctc->count[i] -= (long)subcnt;
				if (ctc->count[i] <= 0L) {
					ctc->int_flag |= bit;
					if (!i) {
						clk0 = 1;
					}
				}
			}
		}

		bit = 1;
		for (i=0; i<4; i++, bit<<=1) {
			if (ctc->int_flag & bit) {
				if (!(ctc->value[i] & 0x80)) {
					ctc->int_flag ^= bit;
				}
				else if (ctcint_flg) {
					ctcint_flg = 0;
					ctc->int_flag ^= bit;
					Z80_Cause_Interrupt(ctc->vector + (BYTE)(i << 1));
				}
			}
		}
	}
}

//----------------------------------------------------------------------

CTC_TABLE *ctctablep(WORD port) {

	port &= 0xfffc;
	if (port == 0x1fa0) {
		return(CTC);
	}
	if (port == 0x1fa8) {
		return(&CTC[1]);
	}
	if ((x1flg.SOUND_SW) && (port == 0x0704)) {
		return(&CTC[2]);
	}
	return(NULL);
}


X1_IOW x1_ctc_w(WORD port, BYTE value) {

	CTC_TABLE	*ctc;
	WORD		p = port;

	if ((ctc = ctctablep(port)) == NULL) {
		return;
	}
	port &= 3;

	if (ctc->value[port] & 0x04) {
		WORD scale = 0;
		long count = 256L;
		ctc->bakval[port] = value;
		if (value) {
			count = (long)value;
		}
		if (!(ctc->value[port] & 0x40)) {
			if (ctc->value[port] & 0x20) {
				scale = 8;
			}
			else {
				scale = 4;
			}
		}
		ctc->countmax[port] = count << scale;
		ctc->count[port] = ctc->countmax[port];
		ctc->value[port] &= (BYTE)~6;
	}
	else if (value & 1) {
		ctc->value[port] = value;
	}
	else {
		if (!port) {								// ver0.25
			ctc->vector = value & 0xf8;
		}
	}
}


X1_IOR x1_ctc_r(WORD port) {

	CTC_TABLE	*ctc;

	if (ctc = ctctablep(port)) {
		port &= 3;
		return(ctc->bakval[port]);
	}
	return(0xff);
}

#endif //T_TUNE
